"""Main entry point of the Command Line Interface for the axon_synthesis package."""

# LICENSE HEADER MANAGED BY add-license-header
#
# Copyright (c) 2023-2024 Blue Brain Project, EPFL.
#
# This file is part of Axon Synthesis.
# See https://github.com/BlueBrain/axon-synthesis for further info.
#
# SPDX-License-Identifier: Apache-2.0
# LICENSE HEADER MANAGED BY add-license-header

import logging
import sys
from pathlib import Path

import click

from axon_synthesis.cli import input_creation
from axon_synthesis.cli import scripts
from axon_synthesis.cli import synthesis
from axon_synthesis.cli import validation
from axon_synthesis.cli.utils import GlobalConfig
from axon_synthesis.cli.utils import configure
from axon_synthesis.utils import setup_logger

seed_option = click.option(
    "--seed",
    type=click.IntRange(min=0),
    default=None,
    help="The random seed.",
)


@click.group(name="axon-synthesis")
@click.version_option()
@click.option(
    "-c",
    "--config",
    type=click.Path(dir_okay=False, exists=True),
    callback=configure,
    is_eager=True,
    expose_value=False,
    show_default=True,
    help="Read option defaults from the specified CFG file.",
)
@click.option(
    "--log-level",
    type=click.Choice(["debug", "info", "warning", "error", "critical"]),
    default="info",
    help="The logger level.",
)
@click.option(
    "-d/-nd",
    "--debug/--no-debug",
    default=False,
    help="Trigger the debug mode.",
)
@seed_option
@click.pass_context
def main(ctx, *_args, **kwargs):
    """A command line tool for axon-synthesis management."""
    debug = kwargs.get("debug", False)
    seed = kwargs.get("seed")
    log_level = kwargs.get("log_level", "info")
    if kwargs.get("debug", False):
        log_level = "debug"

    ctx.ensure_object(GlobalConfig)
    ctx.obj.debug = debug
    ctx.obj.seed = seed
    setup_logger(log_level)

    logger = logging.getLogger()
    logger.info("Running the following command: %s", " ".join(sys.argv))
    logger.info("Running from the following folder: %s", Path.cwd())


main.add_command(synthesis.synthesize)
main.add_command(input_creation.create_inputs)
main.add_command(input_creation.fetch_white_matter_recipe)


@main.group(name="validation")
def validation_group():
    """Subset of commands used to validate axon synthesis."""


validation_group.add_command(validation.mimic)


@main.group(name="scripts")
def scripts_group():
    """Subset of commands used to run some useful scripts for axon synthesis."""


scripts_group.add_command(scripts.random_morphologies)



"""Module to define a base class for relative paths storage and processing."""

# LICENSE HEADER MANAGED BY add-license-header
#
# Copyright (c) 2023-2024 Blue Brain Project, EPFL.
#
# This file is part of Axon Synthesis.
# See https://github.com/BlueBrain/axon-synthesis for further info.
#
# SPDX-License-Identifier: Apache-2.0
#

from copy import deepcopy
from enum import IntEnum
from pathlib import Path
from typing import TYPE_CHECKING
from typing import ClassVar

from axon_synthesis.typing import FileType

FILE_SELECTION = IntEnum("FILE_SELECTION", ["ALL", "REQUIRED_ONLY", "OPTIONAL_ONLY", "NONE"])


class BasePathBuilder:
    """A base class to store relative file paths."""

    _filenames: ClassVar[dict] = {}
    _optional_keys: ClassVar[set[str]] = set()
    _dir_keys: ClassVar[set[str]] = set()

    def __init__(self, path: FileType, *, exists: bool = False, create: bool = False):
        """Create a new BasePathBuilder object.

        Args:
            path: The base path used to build the relative paths.
            exists: If set to True, the given path must already exist.
            create: If set to True, the given path will be automatically created.
        """
        self._filenames = deepcopy(self._filenames)  # type: ignore[misc]
        self._optional_keys = deepcopy(self._optional_keys)  # type: ignore[misc]
        self._dir_keys = deepcopy(self._dir_keys)  # type: ignore[misc]
        self._path = Path(path)
        self._reset_attributes()

        if exists and not self.path.exists():
            msg = f"The directory {self.path} does not exist"
            raise FileNotFoundError(msg)

        if create:
            self.create_root()

    @property
    def path(self) -> Path:
        """Return the associated path."""
        return self._path

    def reset_path(self, path):
        """Return the associated path."""
        self._path = Path(path)
        self._reset_attributes()

    def create_root(self):
        """Ensure the root path exists."""
        self.path.mkdir(parents=True, exist_ok=True)

    def __iter__(self):
        """Return a generator to the paths to the associated data files."""
        yield from self.build_paths(self.path).items()

    def _reset_attributes(self, names=None) -> None:
        """Reset path attributes."""
        if names is None:
            names = self
        for k, v in names:
            setattr(self, k, v)

    if TYPE_CHECKING:

        def __getattr__(self, name: str) -> Path:
            """Custom __getattr__ operator.

            This is only to tell mypy that the dynamically assigned attributes will all return
            Path objects
            """

    def build_paths(self, path) -> dict[str, Path]:
        """Build the paths to the associated data files."""
        return self.build_default_paths(
            path, filenames=self._filenames, optional_keys=self._optional_keys
        )

    @classmethod
    def build_default_paths(cls, path, filenames=None, optional_keys=None) -> dict[str, Path]:
        """Build the default paths to the associated data files of the class."""
        path = Path(path)
        paths = {}
        if filenames is None:
            filenames = cls._filenames
        if optional_keys is None:
            optional_keys = cls._filenames
        for k, v in filenames.items():
            if v is not None:
                paths[k] = path / v
            elif k in optional_keys:
                paths[k] = None
            else:
                msg = f"Only optional keys can be set to None but {k} is not optional"
                raise ValueError(msg)
        return paths

    @property
    def filenames(self):
        """Return the paths to the associated data files."""
        return dict(self)

    @property
    def optional_filenames(self):
        """Return the optional files."""
        return {k: v for k, v in self if k in self._optional_keys}

    @property
    def required_filenames(self):
        """Return the required files."""
        return {k: v for k, v in self if k not in self._optional_keys}

    def filenames_from_type(
        self, *, file_selection: FILE_SELECTION = FILE_SELECTION.ALL
    ) -> dict[str, Path]:
        """Return the associated files according to the given filter tag."""
        if file_selection == FILE_SELECTION.ALL:
            return self.filenames
        if file_selection == FILE_SELECTION.REQUIRED_ONLY:
            return self.required_filenames
        if file_selection == FILE_SELECTION.OPTIONAL_ONLY:
            return self.optional_filenames
        return {}

    def missing_files(self, *, file_selection=FILE_SELECTION.REQUIRED_ONLY):
        """Get the list of missing files based on a given type (required, optional or all)."""
        files = self.filenames_from_type(file_selection=file_selection)
        return {k: v for k, v in files.items() if not v.exists()}

    def exists(self, *, file_selection=FILE_SELECTION.REQUIRED_ONLY):
        """Check if all the paths exist."""
        return not self.missing_files(file_selection=file_selection)

    def raise_missing_files(
        self, *, file_selection=FILE_SELECTION.REQUIRED_ONLY, missing_files=None
    ):
        """Raise a 'FileNotFoundError' exception with the relevant list of missing files."""
        msg = "The following files are missing: %s"
        if missing_files is None:
            missing_files = list(self.missing_files(file_selection=file_selection).values())
        raise FileNotFoundError(msg, missing_files)

    def assert_exists(self, *, file_selection=FILE_SELECTION.REQUIRED_ONLY):
        """Raise a 'FileNotFoundError' exception if the relevant files do not exist."""
        files = list(self.missing_files(file_selection=file_selection).values())
        if files:
            self.raise_missing_files(missing_files=files)

    def is_subdir(self, key):
        """Check if a given key is registered as a subdirectory."""
        return key in self._dir_keys

    def create_dirs(self, *, file_selection: FILE_SELECTION = FILE_SELECTION.REQUIRED_ONLY):
        """Create sub-directories."""
        files = self.filenames_from_type(file_selection=file_selection)
        for k, v in files.items():
            if self.is_subdir(k):
                v.mkdir(parents=True, exist_ok=True)

    def update_from_dict(self, key, attr_name, data):
        """Update an attribute from a given dictionary if the requested value is not None."""
        value = data.get(key, None)
        if value is not None:
            setattr(self, attr_name, value)



"""Helpers for atlas."""

# LICENSE HEADER MANAGED BY add-license-header
#
# Copyright (c) 2023-2024 Blue Brain Project, EPFL.
#
# This file is part of Axon Synthesis.
# See https://github.com/BlueBrain/axon-synthesis for further info.
#
# SPDX-License-Identifier: Apache-2.0
#

import contextlib
import copy
import json
import logging
import operator
from functools import cached_property
from pathlib import Path

import h5py
import numpy as np
import pandas as pd
from attrs import asdict
from attrs import define
from attrs import field
from scipy.ndimage import convolve
from scipy.spatial import KDTree
from voxcell import OrientationField
from voxcell import VoxelData
from voxcell.nexus.voxelbrain import Atlas
from voxcell.voxel_data import ValueToIndexVoxels

from axon_synthesis.typing import ArrayLike
from axon_synthesis.typing import CoordsType
from axon_synthesis.typing import FileType
from axon_synthesis.typing import LayerNamesType
from axon_synthesis.typing import RegionIdsType
from axon_synthesis.typing import SeedType
from axon_synthesis.typing import Self

LOGGER = logging.getLogger(__name__)


def _is_in(test_elements: list, brain_regions: ArrayLike) -> np.ndarray:
    res = np.zeros_like(brain_regions, dtype=bool)
    for i in test_elements:
        res |= brain_regions == i
    return res


@define
class AtlasConfig:
    """Class to store the Atlas configuration.

    Attributes:
        path: The path to the directory containing the atlas.
        region_filename: The name of the file containing the brain regions.
        layer_names: The list of layer names.
        load_region_map: Trigger for region map loading.
        outside_region_id: The brain region ID of the outside of the brain.
        use_boundary: Trigger to use the boundary or not.
    """

    path: Path = field(converter=Path)
    region_filename: Path = field(converter=Path, default=Path("brain_regions"))
    # flatmap_filename: Path = field(converter=Path)
    layer_names: LayerNamesType | None = field(default=None)
    load_region_map: bool = field(default=False)
    outside_region_id: int = field(default=0)
    use_boundary: bool = field(default=True)

    def to_dict(self) -> dict:
        """Return all attribute values into a dictionary."""
        return asdict(self)

    @classmethod
    def from_dict(cls, data: dict) -> Self:
        """Create a new AtlasConfig object from a dictionary."""
        return cls(
            data["path"],
            data["region_filename"],
            # data["flatmap_filename"],
            data.get("layer_names"),
        )


class AtlasHelper:
    """Atlas helper."""

    def __init__(
        self: Self,
        config: AtlasConfig,
        logger=None,
    ):
        """Create a new BasePathBuilder object.

        Args:
            config: The configuration used to load the atlas.
            logger: An optional logger.
        """
        if logger is None:
            self.logger = LOGGER
        else:
            self.logger = logger
        self.config = config

        # Get atlas data
        self.logger.info("Loading atlas with the following config: %s", self.config)
        atlas = Atlas.open(str(self.config.path.resolve()))

        self.logger.debug(
            "Loading brain regions from the atlas using: %s", self.config.region_filename.name
        )
        self.brain_regions = atlas.load_data(self.config.region_filename.stem)
        self.orientations = atlas.load_data("orientation", cls=OrientationField)

        if self.config.load_region_map:
            self.logger.debug("Loading region map from the atlas")
            self.region_map = atlas.load_region_map()
            self._build_region_map_df_level()
        else:
            self.region_map = None

        self.layers = (
            self.config.layer_names if self.config.layer_names is not None else list(range(1, 7))
        )
        self.top_layer = atlas.load_data(f"[PH]{self.layers[0]}")

        # if config.atlas_flatmap_filename is None:
        #     # Create the flatmap of the atlas
        #     logger.debug("Building flatmap")
        #     one_layer_flatmap = np.mgrid[
        #         : brain_regions.raw.shape[2],
        #         : brain_regions.raw.shape[0],
        #     ].T[:, :, ::-1]
        #     flatmap = VoxelData(
        #         np.stack([one_layer_flatmap] * brain_regions.raw.shape[1], axis=1),
        #         voxel_dimensions=brain_regions.voxel_dimensions,
        #     )
        # else:
        #     # Load the flatmap of the atlas
        #     flatmap = atlas.load_data(config.atlas_flatmap_filename)

        # if self.debug_flatmap:
        #     logger.debug(f"Saving flatmap to: {self.output()['flatmap'].path}")
        #     flatmap.save_nrrd(self.output()["flatmap"].path, encoding="raw")

        # TODO: Compute the depth for specific layers of each region (like in region-grower)
        self.y = atlas.load_data("[PH]y")
        self.depths = VoxelData.reduce(operator.sub, [self.pia_coord, self.y])

        if self.config.use_boundary:
            # Load or compute the boundary
            boundary_file = (self.config.path / "boundary").with_suffix(".nrrd")
            if boundary_file.exists():
                self.logger.debug("Loading atlas boundary from the atlas folder")
                self.boundary = atlas.load_data("boundary")
            else:
                self.build_boundary()
                self.logger.debug("Saving atlas boundary to %s", boundary_file)
                self.boundary.save_nrrd(str(boundary_file))
            self.forbidden_regions = self.brain_regions.with_data(
                self.brain_regions.raw == self.config.outside_region_id
            )
        else:
            self.boundary = None
            self.forbidden_regions = None
        self.logger.debug("Atlas loaded")

    def build_boundary(self, logger=None):
        """Build the boundary orientations."""
        if logger is None:
            logger = LOGGER
        self.logger.debug("Computing atlas boundary")

        shape = self.brain_regions.shape
        self.boundary = self.brain_regions.with_data(np.zeros((*shape, 3), dtype=np.float32))

        outside_voxels, _ = self.get_region_voxels(self.config.outside_region_id)
        inside_voxels, _ = self.get_region_voxels(self.config.outside_region_id, inverse=True)
        tree = KDTree(outside_voxels)
        _, nearest_voxel_ids = tree.query(inside_voxels, workers=-1)
        boundary_vectors = (
            inside_voxels - tree.data[nearest_voxel_ids]
        ) * self.brain_regions.voxel_dimensions
        self.boundary.raw[tuple(inside_voxels.T)] = boundary_vectors

        tree = KDTree(inside_voxels)
        _, nearest_voxel_ids = tree.query(outside_voxels, workers=-1)
        boundary_vectors = (
            outside_voxels - tree.data[nearest_voxel_ids]
        ) * self.brain_regions.voxel_dimensions

        # All outside points are considered as very close and going toward the inside of the atlas
        boundary_vectors /= np.repeat(np.linalg.norm(boundary_vectors, axis=-1), 3).reshape(
            boundary_vectors.shape
        )
        boundary_vectors *= -1e-3

        self.boundary.raw[tuple(outside_voxels.T)] = boundary_vectors

        # Smooth the boundary directions
        self.logger.debug("Smoothing atlas boundary")
        kernel_size = np.array([3, 3, 3])
        kernel = np.ones(kernel_size)
        kernel[1, 1, 1] = 2
        kernel /= kernel.sum()

        vec_norm = np.linalg.norm(self.boundary.raw, axis=-1)
        convolved = np.zeros(self.boundary.raw.shape)
        for i in range(3):
            convolved[:, :, :, i] = convolve(self.boundary.raw[:, :, :, i], kernel)

        # Normalize the smoothed field to keep the original distance
        convolved *= np.repeat(vec_norm / np.linalg.norm(convolved, axis=-1), 3).reshape(
            convolved.shape
        )

        self.boundary.raw[:, :, :, :] = convolved[:, :, :, :]

    def save(self, path):
        """Export the atlas to the given directory."""
        path = Path(path)
        path.mkdir(parents=True, exist_ok=True)
        if self.boundary is not None:
            self.boundary.save_nrrd(str(path / "boundary.nrrd"))
        self.brain_regions.save_nrrd(str(path / "brain_regions.nrrd"))
        self.orientations.save_nrrd(str(path / "orientation.nrrd"))
        self.top_layer.save_nrrd(str(path / f"[PH]{self.layers[0]}.nrrd"))
        self.y.save_nrrd(str(path / "[PH]y.nrrd"))
        with (path / "hierarchy.json").open(mode="w", encoding="utf-8") as f:
            json.dump(self.region_map.as_dict(), f, indent=4)

    def copy(self):
        """Return a copy of the current atlas."""
        return copy.deepcopy(self)

    @property
    def region_map(self):
        """Return the region map."""
        return self._region_map

    @region_map.setter
    def region_map(self, value) -> None:
        """Setter for the region map."""
        self._region_map = value
        if value is None:
            self._region_map_df = None
        else:
            self._region_map_df = value.as_dataframe()
            self._build_region_map_df_level()

    @property
    def region_map_df(self):
        """Return the DF representation of the region map."""
        return self._region_map_df

    @cached_property
    def pia_coord(self) -> VoxelData:
        """Return an atlas of the pia coordinate along the principal axis."""
        return self.top_layer.with_data(self.top_layer.raw[..., 1])

    def compute_region_masks(self, output_path: FileType):
        """Compute all region masks."""
        if self.region_map is None:
            msg = "The region_map is not loaded so the region masks can't be computed"
            raise RuntimeError(msg)
        self.logger.info("Computing brain region masks")
        output_path = Path(output_path)

        if output_path.exists():
            self.logger.info(
                "The brain region mask is not computed because it already exists in '%s'",
                output_path,
            )
            return

        output_path.parent.mkdir(parents=True, exist_ok=True)

        region_map_df = (
            self.region_map_df.reset_index()
            .merge(
                self.region_map_df[["acronym"]].reset_index(),
                left_on="parent_id",
                right_on="id",
                suffixes=("", "_parent"),
                how="left",
            )
            .set_index("id")
        )
        region_map_df["self_and_descendants"] = region_map_df.index.to_series().apply(
            lambda row: tuple(sorted(self.region_map.find(row, attr="id", with_descendants=True))),
        )

        # TODO: Maybe we can keep all the masks in memory? It's just a set of lists of ints.
        with h5py.File(output_path, "w") as f:
            index = ValueToIndexVoxels(self.brain_regions.raw)
            for current_id, self_and_descendants_atlas_ids in region_map_df[
                ["self_and_descendants"]
            ].to_records(index=True):
                self.logger.debug("Create mask for %s", current_id)
                coords = index.value_to_indices(self_and_descendants_atlas_ids)
                if len(coords) == 0:
                    self.logger.warning(
                        "No voxel found for atlas ID %s using the following descendants: %s",
                        current_id,
                        self_and_descendants_atlas_ids,
                    )
                f.create_dataset(
                    str(current_id), data=coords, compression="gzip", compression_opts=1
                )

        self.logger.info("Masks exported to %s", output_path)

    def get_region_ids(
        self,
        brain_region_names: RegionIdsType,
        *,
        with_descendants: bool = True,
    ) -> tuple[list[str | int], list[str | int]]:
        """Find brain region IDs from their names, acronyms or direct IDs.

        Args:
            brain_region_names: The names of the brain regions to get IDs.
            with_descendants: If set to True, all the descendants are included.
        """
        if self.region_map is None:
            msg = "The region_map is not loaded so it's not possible to get region IDs"
            raise RuntimeError(msg)

        if isinstance(brain_region_names, int | str):
            brain_region_names = [brain_region_names]
        missing_ids = []
        brain_region_ids = []

        for i in brain_region_names:
            if isinstance(i, str):
                with contextlib.suppress(ValueError):
                    i = int(i)  # noqa: PLW2901

            new_ids = []

            if isinstance(i, int):
                new_ids.extend(
                    list(self.region_map.find(i, attr="id", with_descendants=with_descendants))
                )
            else:
                new_ids.extend(
                    list(self.region_map.find(i, attr="name", with_descendants=with_descendants))
                )
                new_ids.extend(
                    list(
                        self.region_map.find(i, attr="acronym", with_descendants=with_descendants)
                    ),
                )

            if not new_ids:
                missing_ids.append(i)
            else:
                brain_region_ids.extend(new_ids)

        return sorted(set(brain_region_ids)), sorted(set(missing_ids))

    def get_region_voxels(
        self,
        brain_region_names: RegionIdsType,
        *,
        inverse: bool = False,
    ) -> np.ndarray | tuple[np.ndarray, list[str | int]]:
        """Get the coordinates of the voxels located in the given regions from the atlas."""
        brain_region_ids: list[str | int]
        missing_ids: list[str | int]
        brain_region_ids, missing_ids = self.get_region_ids(brain_region_names)

        brain_region_mask = np.isin(
            self.brain_regions.raw, list(set(brain_region_ids).union(missing_ids))
        )
        if inverse:
            brain_region_mask = ~brain_region_mask

        brain_regions_coords = np.argwhere(brain_region_mask)

        return brain_regions_coords, missing_ids

    def get_random_voxel_shifts(self, size, *, rng: SeedType = None):
        """Pick random shifts from voxel centers according to the voxel sizes."""
        rng = np.random.default_rng(rng)
        half_voxels = self.brain_regions.voxel_dimensions / 2
        return rng.uniform(-half_voxels, half_voxels, size=(size, 3))

    def get_region_points(
        self,
        brain_region_names: RegionIdsType,
        *,
        size: int | None = None,
        random_shifts: bool = False,
        inverse: bool = False,
        rng: SeedType = None,
    ) -> np.ndarray | tuple[np.ndarray, list[str | int]]:
        """Extract region points from the atlas.

        If 'random_shifts' is set to False (the default), the voxel centers are returned. If
        'random_shifts' is set to True, one random point inside each of these voxels (chosen using
        the given Random Number Generator) are returned.

        Args:
            brain_region_names: The name of the regions to consider.
            size: The number of points to return (they are chosen randomly along the possible ones).
            random_shifts: If True, a random shift is applied to return a random point inside the
                voxel.
            inverse: If True, choose points that are NOT located in the given regions.
            rng: The random number generator (can be an int used as seed or a numpy Generator).
        """
        if size is not None and size <= 0:
            msg = "The 'size' argument must be a positive integer."
            raise ValueError(msg)

        brain_regions_coords, missing_ids = self.get_region_voxels(
            brain_region_names, inverse=inverse
        )

        voxel_points = self.brain_regions.indices_to_positions(
            brain_regions_coords + [0.5, 0.5, 0.5]  # noqa: RUF005
        )

        if size is not None:
            # Pick the given number of voxels (the voxels are picked before applying shifts so the
            # duplicated voxels have different shifts)
            rng = np.random.default_rng(rng)
            voxel_points = rng.choice(voxel_points, size)

        if random_shifts:
            # Pick a random point inside the voxel
            rng = np.random.default_rng(rng)
            voxel_points += self.get_random_voxel_shifts(len(voxel_points), rng=rng)

        return voxel_points, missing_ids

    def _build_region_map_df_level(self) -> None:
        """Check if region_map_df contains the st_level column and build it if it's missing."""
        if "st_level" in self.region_map_df.columns:
            return
        current_level = -1
        self.region_map_df["st_level"] = current_level
        current_level_ids = self.region_map_df.loc[
            ~self.region_map_df["parent_id"].isin(self.region_map_df.index)
        ]
        while not current_level_ids.empty:
            current_level += 1
            self.region_map_df.loc[current_level_ids.index, "st_level"] = current_level
            current_level_ids = self.region_map_df.loc[
                self.region_map_df["parent_id"].isin(current_level_ids.index)
            ]

    @cached_property
    def brain_regions_and_descendants(self) -> pd.DataFrame:
        """Return the brain regions and their descendants from the hierarchy."""
        if self.region_map is None:
            msg = "The region_map is not loaded so the descendants can't be computed"
            raise RuntimeError(msg)
        return (
            self.region_map_df.index.to_series()
            .apply(
                lambda row: tuple(
                    sorted(self.region_map.find(row, attr="id", with_descendants=True))
                )
            )
            .apply(pd.Series)
            .stack()
            .dropna()
            .astype(int)
            .rename("youth_id")
            .to_frame()
            .merge(self.region_map_df[["st_level"]], left_on="youth_id", right_index=True)
            .reset_index()
            .drop(columns="level_1")
            .sort_values(["id", "st_level"], ascending=[True, False])
            .reset_index(drop=True)
        )

    @cached_property
    def brain_regions_and_ascendants(self) -> pd.DataFrame:
        """Return the brain regions and their ascendants from the hierarchy."""
        if self.region_map is None:
            msg = "The region_map is not loaded so the ascendants can't be computed"
            raise RuntimeError(msg)
        return (
            self.region_map_df.index.to_series()
            .apply(
                lambda row: tuple(sorted(self.region_map.get(row, attr="id", with_ascendants=True)))
            )
            .apply(pd.Series)
            .stack()
            .dropna()
            .astype(int)
            .rename("elder_id")
            .to_frame()
            .merge(self.region_map_df[["st_level"]], left_on="elder_id", right_index=True)
            .reset_index()
            .drop(columns="level_1")
            .sort_values(["id", "st_level"], ascending=[True, False])
            .reset_index(drop=True)
        )

    def get_hemisphere(self, coord: CoordsType, h_axis: int = 2, h_origin: str = "L") -> str:
        """Return the hemisphere of the given coordinate.

        Args:
            coord: The coordinate of the point.
            h_axis: The axis along which to check the hemisphere.
            h_origin: The origin of the hemisphere ('L' or 'R').

        Returns:
            The hemisphere of the given coordinate, 'L' or 'R'.
        """
        hemisphere_frontier = (
            self.brain_regions.bbox[1][h_axis] - self.brain_regions.bbox[0][h_axis]
        ) * 0.5
        if coord[h_axis] < hemisphere_frontier:
            return h_origin
        # else return the opposite hemisphere
        return "R" if h_origin == "L" else "L"

    def place_point_in_hemisphere(
        self, coord: CoordsType, hemisphere: str, h_axis: int = 2, h_origin: str = "L"
    ) -> CoordsType:
        """Place the point in the given hemisphere.

        If the point is already in the given hemisphere, nothing is done.
        If the point is in the opposite hemisphere, an axial symmetry of axis h_axis is performed.

        Args:
            coord: The coordinate of the point.
            hemisphere: The hemisphere where the point should be ('L' or 'R').
            h_axis: The axis along which to check the hemisphere.
            h_origin: The origin of the hemisphere ('L' or 'R').

        Returns:
            The coordinate of the point in the given hemisphere.
        """
        coord_curr_hemisphere = self.get_hemisphere(coord, h_axis, h_origin)
        hemisphere_frontier = (
            self.brain_regions.bbox[1][h_axis] - self.brain_regions.bbox[0][h_axis]
        ) * 0.5
        if (
            coord_curr_hemisphere == "L"
            and hemisphere == "R"
            or coord_curr_hemisphere == "R"
            and hemisphere == "L"
        ):
            coord[h_axis] = 2 * hemisphere_frontier - coord[h_axis]
        # else, the coord is in the correct hemisphere: nothing to do
        return coord



"""Helpers for white matter recipe."""

# LICENSE HEADER MANAGED BY add-license-header
#
# Copyright (c) 2023-2024 Blue Brain Project, EPFL.
#
# This file is part of Axon Synthesis.
# See https://github.com/BlueBrain/axon-synthesis for further info.
#
# SPDX-License-Identifier: Apache-2.0
#

import json
import logging
import shutil
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import ClassVar

import numpy as np
import pandas as pd
import yaml
from attrs import asdict
from attrs import define
from attrs import field
from git import Repo
from scipy.spatial.distance import squareform

from axon_synthesis.atlas import AtlasHelper
from axon_synthesis.base_path_builder import BasePathBuilder
from axon_synthesis.constants import WMR_ATLAS_ID
from axon_synthesis.typing import FileType
from axon_synthesis.typing import Self
from axon_synthesis.utils import cols_from_json
from axon_synthesis.utils import cols_to_json
from axon_synthesis.utils import fill_diag

LOGGER = logging.getLogger(__name__)


def fetch(
    url,
    output_path,
    file_path="white_matter_FULL_RECIPE_v1p20.yaml",
    version_reference=None,
):
    """Fetch the White Natter Recipe file from an internal repository."""
    Path(output_path).parent.mkdir(parents=True, exist_ok=True)
    with TemporaryDirectory() as tmpdir:
        dest = Path(tmpdir) / "tmp_repo"
        Repo.clone_from(url, dest)
        if version_reference is not None:
            r = Repo(dest)
            r.git.checkout(version_reference)
        shutil.copy(dest / file_path, output_path)
    if version_reference is None:
        version_reference = "latest"
    LOGGER.info(
        (
            "Fetched the White Matter Recipe using the '%s' file from the '%s' repository at "
            "version '%s' to the file '%s'"
        ),
        file_path,
        url,
        version_reference,
        output_path,
    )


def load(white_matter_file: FileType) -> dict:
    """Load the white matter recipe from YAML file."""
    white_matter_file = Path(white_matter_file)
    LOGGER.debug("Loading white matter recipe file from: %s", white_matter_file)
    with white_matter_file.open("r", encoding="utf-8") as f:
        return yaml.load(f, Loader=yaml.SafeLoader)


def get_atlas_region_id(region_map, pop_row, col_name, second_col_name=None):
    """Get the ID of an atlas region."""

    def get_ids(region_map, pop_row, col_name) -> tuple[list, str]:
        if not pop_row.isna()[col_name]:
            acronym = pop_row[col_name]
            ids = region_map.find(acronym, attr="acronym")
        else:
            acronym = None
            ids = []
        return ids, acronym

    ids, acronym = get_ids(region_map, pop_row, col_name)
    if len(ids) == 0 and second_col_name is not None:
        ids, new_acronym = get_ids(region_map, pop_row, second_col_name)
        if len(ids) == 1 and acronym is not None:
            LOGGER.warning(
                "Could not find any ID for %s in the region map but found one for %s",
                acronym,
                new_acronym,
            )
    else:
        new_acronym = None

    if len(ids) > 1:
        msg = (
            f"Found several IDs for the acronym '{acronym or new_acronym}' in the region "
            f"map: {sorted(ids)}"
        )
        raise ValueError(msg)
    if len(ids) == 0:
        msg = f"Could not find the acronym '{acronym or new_acronym}' in the region map"
        raise ValueError(msg)
    return ids.pop()


@define
class WmrConfig:
    """Class to store the WhiteMatterRecipe configuration."""

    path: Path = field(converter=Path)
    subregion_uppercase: bool
    subregion_remove_prefix: bool
    sub_region_separator: str

    def to_dict(self) -> dict:
        """Return all attribute values into a dictionary."""
        return asdict(self)

    @classmethod
    def from_dict(cls, data: dict) -> Self:
        """Create a new AtlasConfig object from a dictionary."""
        return cls(
            data["path"],
            data["subregion_uppercase"],
            data["subregion_remove_prefix"],
            data["sub_region_separator"],
        )


class WhiteMatterRecipe(BasePathBuilder):
    """Class to store the White Matter Recipe data."""

    _filenames: ClassVar[dict] = {
        "FRACTIONS_FILENAME": "fractions.json",
        "INTERACTION_STRENGTHS_FILENAME": "interaction_strengths.json",
        "LAYER_PROFILES_FILENAME": "layer_profiles.csv",
        "POPULATIONS_FILENAME": "populations.csv",
        "PROJECTION_TARGETS_FILENAME": "projection_targets.csv",
        "PROJECTIONS_FILENAME": "projections.csv",
        "REGION_DATA_FILENAME": "region_data.csv",
        "TARGETS_FILENAME": "targets.csv",
    }

    def __init__(
        self,
        path,
        *,
        load=True,
    ):  # pylint: disable=redefined-outer-name
        """Create a new WhiteMatterRecipe object.

        Args:
            path: The base path used to build the relative paths.
            load: If set to False the internal data are not automatically loaded.
        """
        super().__init__(path)

        self.populations: pd.DataFrame | None = None
        self.projections: pd.DataFrame | None = None
        self.targets: pd.DataFrame | None = None
        self.fractions: dict | None = None
        self.interaction_strengths: dict | None = None
        self.projection_targets: pd.DataFrame | None = None
        self.layer_profiles: pd.DataFrame | None = None
        self.region_data: pd.DataFrame | None = None

        if load:
            self.assert_exists()
            self.load()

    def save(self):
        """Save the White Matter Recipe into the given directory."""
        self.path.mkdir(parents=True, exist_ok=True)

        # Export the population DataFrame
        if self.populations is not None:
            populations = cols_to_json(self.populations, ["atlas_region", "filters"])
            populations.to_csv(self.POPULATIONS_FILENAME, index=False)

        # Export the projection DataFrame
        if self.projections is not None:
            projections = cols_to_json(
                self.projections,
                ["mapping_coordinate_system", "targets", "atlas_region", "filters"],
            )
            projections.to_csv(self.PROJECTIONS_FILENAME, index=False)

        # Export the targets DataFrame
        if self.targets is not None:
            targets = cols_to_json(self.targets, ["target"])
            targets.to_csv(self.TARGETS_FILENAME, index=False)

        # Export the projection DataFrame
        if self.projection_targets is not None:
            projection_targets = cols_to_json(
                self.projection_targets,
                ["targets", "atlas_region", "filters", "target", "topographical_mapping"],
            )
            projection_targets.to_csv(self.PROJECTION_TARGETS_FILENAME, index=False)

        # Export the fractions
        if self.fractions is not None:
            with (self.FRACTIONS_FILENAME).open("w", encoding="utf-8") as f:
                json.dump(self.fractions, f, indent=4, sort_keys=True)

        # Export the interaction strengths
        if self.interaction_strengths is not None:
            with (self.INTERACTION_STRENGTHS_FILENAME).open("w", encoding="utf-8") as f:
                json.dump(
                    {k: v.to_dict("index") for k, v in self.interaction_strengths.items()},
                    f,
                    indent=4,
                    sort_keys=True,
                )

        # Export the layer profiles
        if self.layer_profiles is not None:
            layer_profiles = cols_to_json(self.layer_profiles, ["layers"])
            layer_profiles.to_csv(self.LAYER_PROFILES_FILENAME, index=False)

        # Export the region data
        if self.region_data is not None:
            self.region_data.to_csv(self.REGION_DATA_FILENAME, index=False)

    def load(self):
        """Load the White Matter Recipe from the associated directory."""
        LOGGER.info(
            "Loading the White Matter Recipe from '%s'",
            self.path,
        )
        populations = pd.read_csv(self.POPULATIONS_FILENAME)
        self.populations = cols_from_json(populations, ["atlas_region", "filters"])

        projections = pd.read_csv(self.PROJECTIONS_FILENAME)
        self.projections = cols_from_json(
            projections,
            ["mapping_coordinate_system", "targets", "atlas_region", "filters"],
        )

        targets = pd.read_csv(self.TARGETS_FILENAME)
        self.targets = cols_from_json(targets, ["target"])

        with self.FRACTIONS_FILENAME.open("r", encoding="utf-8") as f:
            self.fractions = json.load(f)

        with self.INTERACTION_STRENGTHS_FILENAME.open("r", encoding="utf-8") as f:
            self.interaction_strengths = json.load(f)

        projection_targets = pd.read_csv(self.PROJECTION_TARGETS_FILENAME)
        self.projection_targets = cols_from_json(
            projection_targets,
            ["targets", "atlas_region", "filters", "target", "topographical_mapping"],
        )

        layer_profiles = pd.read_csv(self.LAYER_PROFILES_FILENAME)
        self.layer_profiles = cols_from_json(layer_profiles, ["layers"])

        self.region_data = pd.read_csv(self.REGION_DATA_FILENAME)

    def compute_probabilities(self, atlas: AtlasHelper):
        """Compute projection probabilities from the White Matter Recipe."""
        LOGGER.info("Computing the probabilities from the white matter")

        if self.projection_targets is None:
            msg = (
                "The 'projection_targets' DataFrame is not loaded yet so it is not possible to "
                "compute the probabilities"
            )
            raise RuntimeError(msg)

        projection_targets = self.projection_targets.loc[
            ~self.projection_targets["target_region_atlas_id"].isna()
        ]
        projection_targets = projection_targets.fillna(
            {"target_subregion_atlas_id": projection_targets["target_region_atlas_id"]},
        ).astype({"target_region_atlas_id": int, "target_subregion_atlas_id": int})

        # Get the projection matrix
        projection_matrix = (
            pd.DataFrame.from_records(self.fractions)  # type: ignore[call-overload]
            .stack()
            .rename("target_projection_strength")
            .reset_index()
            .rename(columns={"level_0": "target_projection_name", "level_1": "pop_raw_name"})
        )

        all_targets = projection_targets.merge(
            projection_matrix,
            on=["pop_raw_name", "target_projection_name"],
            how="left",
        )
        all_targets["probability"] = (
            all_targets["target_projection_strength"]
            * all_targets["target_layer_profile_region_prob"]
        )
        all_targets["subregion_acronym"] = all_targets.fillna(
            {"subregion_acronym": all_targets["region_acronym"]}, inplace=True
        )
        all_targets = all_targets.rename(
            columns={"atlas_region_id": "source_brain_region_id", "pop_raw_name": "population_id"}
        )

        all_targets["target_brain_region_acronym"] = all_targets["target_subregion_acronym"].fillna(
            all_targets["target_region"]
        )
        all_targets["target_brain_region_id"] = all_targets["target_brain_region_acronym"].apply(
            lambda x: atlas.get_region_ids(x, with_descendants=False)[0][0]
        )

        population_probabilities = (
            all_targets[["source_brain_region_id", "population_id"]]
            .drop_duplicates()
            .rename(columns={"source_brain_region_id": "brain_region_id"})
        )
        population_probabilities = population_probabilities.merge(
            (1 / population_probabilities.groupby(["brain_region_id"]).size()).rename(  # type: ignore[call-overload]
                "probability"
            ),
            left_on="brain_region_id",
            right_index=True,
        )

        projection_probabilities = (
            all_targets[
                [
                    "source_brain_region_id",
                    "population_id",
                    "target_brain_region_id",
                    "target_projection_name",
                    "probability",
                ]
            ]
            .drop_duplicates()
            .rename(
                columns={
                    "population_id": "source_population_id",
                    "target_projection_name": "target_population_id",
                }
            )
        )
        projection_probabilities = projection_probabilities.loc[
            projection_probabilities["probability"] > 0
        ]

        return population_probabilities, projection_probabilities

    def load_from_raw_wmr(  # noqa: PLR0915
        self,
        config: WmrConfig,
        atlas: AtlasHelper,
    ):
        """Process the white matter recipe."""
        # pylint: disable=too-many-statements
        LOGGER.info("Loading and processing the white matter recipe YAML file '%s'", config.path)
        region_map = atlas.region_map
        brain_regions = atlas.brain_regions
        wmr = load(config.path)

        # Get populations
        LOGGER.debug("Extracting populations from white matter recipe")
        wm_populations = pd.DataFrame.from_records(wmr["populations"])
        wm_populations_sub = wm_populations.loc[
            wm_populations["atlas_region"].apply(lambda x: isinstance(x, list)),
            "atlas_region",
        ]
        if not wm_populations_sub.empty:
            wm_populations_sub = (
                wm_populations_sub.apply(pd.Series)  # type: ignore[call-overload]
                .stack()
                .dropna()
                .rename("atlas_region_split")
                .reset_index(level=1, drop=True)
            )
            wm_populations = wm_populations.join(wm_populations_sub, how="left")
            wm_populations["atlas_region_split"].fillna(
                wm_populations["atlas_region"],
                inplace=True,
            )
            wm_populations.drop(columns=["atlas_region"], inplace=True)
            wm_populations.rename(columns={"atlas_region_split": "atlas_region"}, inplace=True)
        wm_populations.rename(columns={"name": "pop_raw_name"}, inplace=True)
        wm_populations["region_acronym"] = wm_populations["atlas_region"].apply(
            lambda row: row["name"],
        )
        wm_populations_sub = (
            wm_populations["atlas_region"]  # type: ignore[call-overload]
            .apply(lambda row: pd.Series(row.get("subregions", []), dtype=object))
            .stack()
            .dropna()
            .rename("sub_region")
            .reset_index(level=1, drop=True)
        )
        wm_populations = wm_populations.join(wm_populations_sub, how="left")

        # Get subregion names
        wm_populations["formatted_subregion"] = wm_populations["sub_region"]
        if config.subregion_uppercase:
            wm_populations["formatted_subregion"] = wm_populations[
                "formatted_subregion"
            ].str.upper()
        if config.subregion_remove_prefix:
            wm_populations["formatted_subregion"] = wm_populations[
                "formatted_subregion"
            ].str.extract(r"(\d+.*)")
        wm_populations["subregion_acronym"] = (
            wm_populations["region_acronym"]
            + config.sub_region_separator
            + wm_populations["formatted_subregion"]
        )

        # Get atlas subregion IDs
        wm_populations["atlas_region_id"] = wm_populations.apply(
            lambda row: get_atlas_region_id(region_map, row, "subregion_acronym", "region_acronym"),
            axis=1,
        )

        # Compute the volume of each region
        region_ids, region_counts = np.unique(brain_regions.raw, return_counts=True)
        region_data = pd.DataFrame({"atlas_region_id": region_ids, "count": region_counts})

        # Get upper-level regions from the populations
        region_data = region_data.merge(
            wm_populations[["atlas_region_id"]],
            on="atlas_region_id",
            how="outer",
        )
        region_data.drop_duplicates(inplace=True)

        # Compute the volumes of upper-level regions
        # TODO: Check if the attr should be 'id' or 'atlas_id'
        region_data["nb_voxels"] = region_data["atlas_region_id"].apply(
            lambda row: region_counts[
                np.argwhere(
                    np.isin(
                        region_ids,
                        [
                            region_map.get(i, WMR_ATLAS_ID)
                            for i in region_map.find(row, attr=WMR_ATLAS_ID, with_descendants=True)
                        ],
                    ),
                )
            ].sum(),
        )
        region_data["atlas_region_volume"] = region_data["nb_voxels"] * brain_regions.voxel_volume
        region_data.drop(columns=["count"], inplace=True)

        # Join region data to population
        wm_populations = wm_populations.merge(region_data, on="atlas_region_id", how="left")

        # Compute volume fractions of sub-regions
        region_ids = wm_populations.apply(
            lambda row: get_atlas_region_id(region_map, row, "region_acronym"),
            axis=1,
        ).rename("region_id")
        pop_frac = wm_populations.join(region_ids)
        pop_frac = pop_frac.merge(
            region_data,
            left_on="region_id",
            right_on="atlas_region_id",
            how="left",
            suffixes=("", "_total"),
        )
        wm_populations["sub_region_volume_frac"] = (
            (pop_frac["atlas_region_volume"] / pop_frac["atlas_region_volume_total"])
            .fillna(1)
            .clip(0, 1)
        )

        # Get layer_profiles
        LOGGER.debug("Extracting layer profiles from white matter recipe")
        wm_layer_profiles = pd.DataFrame.from_records(wmr["layer_profiles"])
        layer_profiles = (
            wm_layer_profiles["relative_densities"]  # type: ignore[call-overload]
            .apply(pd.Series)
            .stack()
            .rename("layer_profile")
            .reset_index(level=1)
            .rename(columns={"level_1": "layer_profile_num"})
        )
        wm_layer_profiles = wm_layer_profiles.join(layer_profiles).set_index(
            "layer_profile_num",
            append=True,
        )
        wm_layer_profiles["layers"] = wm_layer_profiles["layer_profile"].apply(
            lambda row: row.get("layers", None),
        )
        wm_layer_profiles["value"] = wm_layer_profiles["layer_profile"].apply(
            lambda row: row.get("value", None),
        )
        wm_layer_profiles.drop(columns=["relative_densities", "layer_profile"], inplace=True)
        wm_layer_profiles = wm_layer_profiles.join(
            wm_layer_profiles["layers"]  # type: ignore[call-overload]
            .apply(pd.Series)
            .stack()
            .rename("layer")
            .reset_index(level=2)
            .rename(columns={"level_2": "layer_index"}),
        )
        wm_layer_profiles["formatted_layer"] = wm_layer_profiles["layer"].str.extract("l(.*)")
        wm_layer_profiles.fillna({"formatted_layer": wm_layer_profiles["layer"]}, inplace=True)

        # Get projections
        LOGGER.debug("Extracting projections from white matter recipe")
        wm_projections = pd.DataFrame.from_records(wmr["projections"])
        if wm_projections["source"].duplicated().any():
            msg = (
                "Found several equal sources in the 'projections' entry: "
                f"{sorted(wm_projections.loc[wm_projections['a'].duplicated(), 'a'].tolist())}",
            )
            raise ValueError(msg)

        # Map projections
        wm_projections = wm_projections.merge(
            wm_populations,
            left_on="source",
            right_on="pop_raw_name",
            how="left",
        )

        wm_targets = (
            wm_projections["targets"]  # type: ignore[call-overload]
            .apply(pd.Series)
            .stack()
            .rename("target")
            .reset_index(level=1)
            .rename(columns={"level_1": "target_num"})
        )
        wm_projection_targets = wm_projections.join(wm_targets).set_index("target_num", append=True)
        wm_projection_targets["target_population_name"] = wm_projection_targets["target"].apply(
            lambda row: row["population"],
        )
        wm_projection_targets["target_projection_name"] = wm_projection_targets["target"].apply(
            lambda row: row["projection_name"],
        )
        wm_projection_targets["target_density"] = wm_projection_targets["target"].apply(
            lambda row: row["density"],
        )
        wm_projection_targets["topographical_mapping"] = wm_projection_targets["target"].apply(
            lambda row: row["presynaptic_mapping"],
        )
        wm_projection_targets["target_layer_profiles"] = wm_projection_targets["target"].apply(
            lambda row: row["target_layer_profiles"],
        )
        wm_projection_targets.index.rename("proj_index", level=0, inplace=True)  # type: ignore[call-arg,call-overload]

        # Get target sub regions
        region_map_df = region_map.as_dataframe()
        if "atlas_id" not in region_map_df.columns:
            region_map_df["atlas_id"] = region_map_df.index.to_numpy()
        region_map_df = (
            region_map_df.reset_index()
            .merge(
                region_map_df[["acronym"]].reset_index(),
                left_on="parent_id",
                right_on="id",
                suffixes=("", "_parent"),
                how="left",
            )
            .set_index("id")
        )
        sub_region_acronyms = region_map_df.groupby(["acronym_parent"])["acronym"].apply(list)
        sub_region_acronyms.index.rename("region_acronym", inplace=True)
        sub_region_acronyms.rename("subregion_acronyms", inplace=True)
        # sub_region_atlas_ids = region_map_df.groupby(["acronym_parent"])["atlas_id"].apply(list)
        # sub_region_atlas_ids.index.rename("region_acronym", inplace=True)
        # sub_region_atlas_ids.rename("subregion_atlas_ids", inplace=True)

        # Join layer profiles

        # Stack sub-regions

        target_pop_name = (
            wm_projection_targets["target_population_name"]
            .str.extract("(.*)(_ALL_LAYERS)?")[0]
            .rename("target_pop_name")
            .to_frame()
        )
        wm_projection_targets["target_region"] = (
            target_pop_name.reset_index()
            .merge(
                wm_populations[["pop_raw_name", "region_acronym"]],
                left_on="target_pop_name",
                right_on="pop_raw_name",
            )
            .set_index(["proj_index", "target_num"])["region_acronym"]
            .sort_index()
        )
        wm_projection_targets.fillna(
            {"target_region": wm_projection_targets["target_population_name"]},
            inplace=True,
        )
        wm_projection_targets = (
            wm_projection_targets.reset_index()
            .merge(
                region_map_df[["acronym", "atlas_id"]],
                left_on="target_region",
                right_on="acronym",
                how="left",
            )
            .set_index(["proj_index", "target_num"])
            .rename(columns={"atlas_id": "target_region_atlas_id"})
            .drop(columns=["acronym"])
        )
        wm_projection_targets = wm_projection_targets.merge(
            sub_region_acronyms,
            left_on="target_region",
            right_index=True,
            how="left",
        )

        selected_sub_region_acronyms = (
            wm_projection_targets["subregion_acronyms"]  # type: ignore[call-overload]
            .apply(pd.Series)
            .stack()
            .rename("subregion_acronym")
            .reset_index(level=2)
            .rename(columns={"level_2": "subregion_num"})
        )
        wm_projection_targets = wm_projection_targets.join(
            selected_sub_region_acronyms,
            rsuffix="_target",
        ).rename(columns={"subregion_acronym_target": "target_subregion_acronym"})
        wm_projection_targets = (
            wm_projection_targets.reset_index()
            .merge(
                region_map_df[["acronym", "atlas_id"]],
                left_on="target_subregion_acronym",
                right_on="acronym",
                how="left",
            )
            .set_index(["proj_index", "target_num"])
            .rename(columns={"atlas_id": "target_subregion_atlas_id"})
            .drop(columns=["acronym"])
        )

        # selected_sub_region_acronyms.index.rename("index", level=0, inplace=True)

        # Drop target sub-regions that are not listed in the populations
        wm_projection_targets = wm_projection_targets.loc[
            wm_projection_targets["target_subregion_acronym"].isin(
                wm_populations["subregion_acronym"].drop_duplicates(),
            )
        ]

        # Join layer profiles
        wm_projection_targets = wm_projection_targets.merge(
            wm_populations[
                ["subregion_acronym", "sub_region", "formatted_subregion", "sub_region_volume_frac"]
            ].drop_duplicates(),
            left_on="target_subregion_acronym",
            right_on="subregion_acronym",
            how="left",
            suffixes=("", "_target"),
        )
        wm_projection_targets["target_layer_profile_name"] = wm_projection_targets.apply(
            lambda row: row["target_layer_profiles"][0].get("name"),
            axis=1,
        )
        # wm_projection_targets.drop(columns=["targets", "target_layer_profiles"], inplace=True)
        wm_projection_targets = (
            wm_projection_targets.reset_index()
            .merge(
                wm_layer_profiles[["name", "layer", "formatted_layer", "value"]].rename(
                    columns={"value": "target_layer_profile_density"},
                ),
                left_on=["target_layer_profile_name", "sub_region_target"],
                right_on=["name", "layer"],
                how="left",
            )
            .set_index("index")
        )
        wm_projection_targets.fillna({"target_layer_profile_density": 1}, inplace=True)
        wm_projection_targets["target_layer_profile_prob"] = (
            wm_projection_targets["target_layer_profile_density"]
            * wm_projection_targets["sub_region_volume_frac_target"]
        )

        normalization_factor = (
            wm_projection_targets.groupby(
                ["source", "sub_region", "target_projection_name", "target_region"],
            )["target_layer_profile_prob"]
            .sum()
            .rename("target_layer_profile_norm_factor")
        )
        wm_projection_targets = wm_projection_targets.merge(
            normalization_factor,
            left_on=["source", "sub_region", "target_projection_name", "target_region"],
            right_index=True,
            how="left",
        )
        wm_projection_targets.fillna({"target_layer_profile_norm_factor": 1}, inplace=True)
        wm_projection_targets["target_layer_profile_region_prob"] = (
            wm_projection_targets["target_layer_profile_prob"]
            / wm_projection_targets["target_layer_profile_norm_factor"]
        )
        # wm_projection_targets["partial_strength"] = (
        #     wm_projection_targets["strength"]
        #     * wm_projection_targets["target_layer_profile_norm_factor"]
        # )

        # Ignore the sub-regions not listed in layer profiles

        # wm_projection_targets["target_region"].fillna("target_population_name", inplace=True)

        # target_all_layers_mask = wm_projection_targets["target_population_name"].str.endswith(
        #     "_ALL_LAYERS"
        # )
        # wm_projection_targets["target_sub_regions"] = None
        # wm_projection_targets.loc[target_all_layers_mask, "target_sub_regions"]

        # Use region ID when there is no sub-region
        wm_projection_targets["target_atlas_id"] = wm_projection_targets[
            "target_subregion_atlas_id"
        ].fillna(wm_projection_targets["target_region_atlas_id"])
        wm_projection_targets.fillna({"target_layer_profile_region_prob": 1}, inplace=True)
        wm_projection_targets["has_atlas_id"] = ~wm_projection_targets[
            "target_subregion_atlas_id"
        ].isna()

        # # Compute normalization factors

        # # Compute final probabilities

        # wm_projection_targets.merge(
        #     wm_populations[
        #         [
        #             "pop_raw_name",
        #             "region_acronym",
        #             "sub_region",
        #             "formatted_subregion",
        #             "subregion_acronym",
        #             "atlas_region_id",
        #             "atlas_region_volume",
        #             "sub_region_volume_frac",
        #         ]
        #     ],
        #     left_on="target_population_name",
        #     right_on="pop_raw_name",
        #     how="left",
        #     suffixes=("_src", "_trgt"),
        # )

        # wm_projection_targets["layer_profile_name"] = wm_projection_targets.apply(
        #     lambda row: row["target_layer_profiles"][0].get("name"), axis=1
        # )
        # wm_projection_targets.drop(columns=["targets", "target_layer_profiles"], inplace=True)
        # wm_projection_targets = (
        #     wm_projection_targets.reset_index()
        #     .merge(
        #         wm_layer_profiles[["name", "layer", "formatted_layer", "value"]].rename(
        #             columns={"value": "layer_profile_density"}
        #         ),
        #         left_on=["layer_profile_name", "sub_region"],
        #         right_on=["name", "layer"],
        #         how="left",
        #     )
        #     .set_index(["level_0", "target_num"])
        # )
        # wm_projection_targets["layer_profile_density"].fillna(1, inplace=True)

        # wm_projection_targets["layer_profile_prob"] = (
        #     wm_projection_targets["layer_profile_density"]
        #     * wm_projection_targets["sub_region_volume_frac"]
        # )

        # print()
        # wm_projection_targets.groupby("region_acronym")["layer_profile_prob"].sum()

        # wm_projection_targets.drop(columns=["atlas_region", "name", "layer"], inplace=True)

        # for i in target.get("target_layer_profiles", []):
        #     layer_fraction = i.get("fraction", 1.0)
        #     layer_name = i["name"]

        # Get fractions
        LOGGER.debug("Extracting fractions from white matter recipe")
        wm_fractions = {i["population"]: i["fractions"] for i in wmr["p-types"]}

        # Get interaction_mat and strengths
        LOGGER.debug("Extracting interaction matrix from white matter recipe")
        wm_interaction_mat = {
            i["population"]: i["interaction_mat"] for i in wmr["p-types"] if "interaction_mat" in i
        }

        wm_interaction_strengths = {
            k: pd.DataFrame(
                fill_diag(squareform(v["strengths"]), 1),
                columns=wm_interaction_mat[k]["projections"],
                index=wm_interaction_mat[k]["projections"],
            )
            for k, v in wm_interaction_mat.items()
        }

        self.populations = wm_populations
        self.projections = wm_projections
        self.targets = wm_targets
        self.fractions = wm_fractions
        self.interaction_strengths = wm_interaction_strengths
        self.projection_targets = wm_projection_targets
        self.layer_profiles = wm_layer_profiles
        self.region_data = region_data



"""Module with some validation utils."""

# LICENSE HEADER MANAGED BY add-license-header
#
# Copyright (c) 2023-2024 Blue Brain Project, EPFL.
#
# This file is part of Axon Synthesis.
# See https://github.com/BlueBrain/axon-synthesis for further info.
#
# SPDX-License-Identifier: Apache-2.0
#

from pathlib import Path

import numpy as np
import pandas as pd
from morph_tool.converter import convert
from voxcell import VoxelData
from voxcell.math_utils import voxel_intersection

from axon_synthesis.constants import FROM_COORDS_COLS
from axon_synthesis.constants import TO_COORDS_COLS
from axon_synthesis.typing import FileType
from axon_synthesis.utils import disable_loggers
from axon_synthesis.utils import temp_dir


def segment_voxel_intersections(row, grid, *, return_sub_segments=False):
    """Get indices and intersection lengths of the voxels intersected by the given segment."""
    start_pt = [row[FROM_COORDS_COLS.X], row[FROM_COORDS_COLS.Y], row[FROM_COORDS_COLS.Z]]
    end_pt = [row[TO_COORDS_COLS.X], row[TO_COORDS_COLS.Y], row[TO_COORDS_COLS.Z]]
    indices, sub_segments = voxel_intersection(
        [start_pt, end_pt],
        grid,
        return_sub_segments=True,
    )
    res = {
        "indices": indices,
        "lengths": np.linalg.norm(sub_segments[:, [3, 4, 5]] - sub_segments[:, [0, 1, 2]], axis=1),
    }
    if return_sub_segments:
        res["sub_segments"] = sub_segments
    return pd.Series(res)


def segment_intersection_lengths(segments, bbox, voxel_dimensions, center=None, logger=None):
    """Compute the intersection lengths of the given segments with the given grid."""
    shape = np.clip((bbox[1] - bbox[0]) // voxel_dimensions, 1, np.inf).astype(int)
    if center is not None:
        shape += 3
    if logger is not None:
        logger.debug(
            "Create grid with size=%s, voxel_dimensions=%s and offset=%s",
            shape,
            voxel_dimensions,
            bbox[0],
        )
    grid = VoxelData(np.zeros(shape), voxel_dimensions, offset=bbox[0])

    # Ensure the center is located at the center of a voxel
    if center is not None:
        grid.offset -= (
            1.5 - np.modf(grid.positions_to_indices(center, keep_fraction=True))[0]
        ) * voxel_dimensions

    # Compute intersections
    intersections = segments.apply(segment_voxel_intersections, args=(grid,), axis=1)

    elements = pd.DataFrame(
        {
            "indices": intersections["indices"].explode(),
            "lengths": intersections["lengths"].explode(),
        }
    )
    outside_segments = elements.loc[elements["indices"].isna()]
    if not outside_segments.empty:
        if logger is not None:
            logger.warning("Found %s segments outside the given grid", len(outside_segments))
        elements = elements.dropna(subset=["indices"])
    elements["indices"] = elements["indices"].apply(tuple)

    lengths = elements.groupby("indices")["lengths"].sum().reset_index()
    indices = tuple(np.vstack(lengths["indices"].to_numpy()).T.tolist())  # type: ignore[call-overload]

    grid.raw[indices] += lengths["lengths"].astype(float).to_numpy()
    return grid


def copy_morph_to_tmp_dir(morph):
    """Copy the given morphology in a temporary directory."""
    tmp_dir = temp_dir()
    if isinstance(morph, FileType):  # type: ignore[arg-type,misc]
        filename = Path(morph)
        name = filename.stem
        ext = filename.suffix
    else:
        name = morph.name
        ext = ".asc"
    filename = (Path(tmp_dir.name) / name).with_suffix(ext)
    with disable_loggers("morph_tool.converter"):
        convert(morph, filename, nrn_order=True)
    return tmp_dir, filename
